# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2016/09/12
# copy: (C) Copyright 2016-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import re

#------------------------------------------------------------------------------
class Engine(object):

  mimetype              = None
  extensions            = None
  open_markers          = ('<',)
  close_markers         = ('>',)

  #----------------------------------------------------------------------------
  def __init__(self, *args, **kw):
    super(Engine, self).__init__(*args, **kw)
    self._ws_cre = re.compile(
      '(?P<close>' + '|'.join([re.escape(m) for m in self.close_markers]) + ')'
      + '\\s*\n\\s*'
      + '(?P<open>' + '|'.join([re.escape(m) for m in self.open_markers]) + ')',
      flags = re.DOTALL)

  #----------------------------------------------------------------------------
  def whitespace(self, text, attrs):
    '''
    Remove "ignorable" whitespace from `text`. Note that `text` will
    already have been dedented and stripped.
    '''
    # todo: ok, this is a *SUPER* simple ignorable whitespace
    #       detection algorithm... this needs to be refactored
    #       somehow...
    #return text
    return self._ws_cre.sub('\\g<close>\\g<open>', text)

  #----------------------------------------------------------------------------
  def precompile(self, text, attrs):
    '''
    Pre-compiles the JavaScript template `text` for delivery to a
    JavaScript enabled client in a "pre-compiled"
    format. Pre-compilation means that the template character stream
    has been, at minimum, parsed into tokens and typically rendered
    into JavaScript native syntax in order to accelerate template
    parsing on the client via highly optimized JavaScript parsing
    code. If pre-compilation is not available (e.g. because the
    specified template type does not support pre-compilation or the
    necessary pre-requisites for pre-compilation are not available),
    then a `jstc.api.PrecompilerUnavailable` exception (or subclass)
    should be raised. The `attrs` specifies template attributes; see
    `jstc.compiler.Compiler.compile` for details.
    '''
    raise NotImplementedError()


#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2016/09/12
# copy: (C) Copyright 2016-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import subprocess
import errno
import json

import asset

from . import base
from .. import api
from .. import compiler

#------------------------------------------------------------------------------
@asset.plugin('jstc.engines.plugins', 'text/x-handlebars')
class HandlebarsEngine(base.Engine):

  mimetype              = 'text/x-handlebars'
  extensions            = ('.hbs', '.handlebars')
  open_markers          = base.Engine.open_markers + ('{{',)
  close_markers         = base.Engine.close_markers + ('}}',)
  executable            = 'handlebars'
  parameters            = ()
  softfail              = ('ENOENT',)

  #----------------------------------------------------------------------------
  def __init__(self, executable=None, parameters=None, softfail=None, *args, **kw):
    super(HandlebarsEngine, self).__init__(*args, **kw)
    if executable is not None:
      self.executable = executable
    if parameters is not None:
      self.parameters = tuple(parameters)
    if softfail is not None:
      self.softfail = tuple(softfail)

    # todo: verify `handlebars --version >= 4.0.5`, or if `handlebars`
    #       is not available, cache that information...? prolly not,
    #       that way a server does not need to be restarted if it is
    #       installed (it'll "just start working") -- let pyramid_jitt
    #       or webassets deal with that.

  #----------------------------------------------------------------------------
  def precompile(self, text, attrs):
    args = [self.executable] + list(self.parameters) + ['--string', text, '--simple']
    if attrs.get('name'):
      args += ['--name', attrs.get('name')]
    res = (compiler.Compiler.TYPE_JSFRAG, self.execute(args).strip())
    if attrs.get('partial'):
      return res
    return self.assemble([res + (attrs,)])

  #----------------------------------------------------------------------------
  def execute(self, args, input=None):
    # todo: move this into an ``ExternalPipeEngine`` helper superclass...
    try:
      if input:
        # todo: implement
        raise NotImplementedError()
      return subprocess.check_output(args)
    except OSError as err:
      if err.errno == errno.ENOENT and 'ENOENT' in self.softfail:
        raise api.PrecompilerUnavailable()
      raise

  #----------------------------------------------------------------------------
  def assemble(self, parts):
    ret = '(function(){var t=Handlebars.template,ts=Handlebars.templates=Handlebars.templates||{};'
    for part in parts:
      contentType, content, attrs = part[:3]
      # todo: confirm correct contentType & content?...
      # todo: confirm attrs.name exists?...
      ret += 'ts[{name}]=t({content});'.format(
        name    = json.dumps(attrs.get('name')),
        content = content,
      )
    ret += '})();'
    return (compiler.Compiler.TYPE_JS, ret)

#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2016/09/12
# copy: (C) Copyright 2016-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import logging
import os.path
import shlex
import re
import textwrap
import cgi

import six
import asset
import morph
from aadict import aadict

from . import api

#------------------------------------------------------------------------------

log = logging.getLogger(__name__)

#------------------------------------------------------------------------------
class TemplateAttributeParser(shlex.shlex):
  # todo: this currently supports ``name; !flag`` but not ``name; ! flag``...
  #       make that work too.
  TOKEN_KEYEND  = ':'
  TOKEN_VALEND  = ';'
  TOKENS        = (TOKEN_KEYEND, TOKEN_VALEND)
  def __init__(self, instream, *args, **kw):
    kw['posix'] = kw.get('posix', True)
    self.src = None
    if morph.isstr(instream):
      self.src = instream
    shlex.shlex.__init__(self, instream, *args, **kw)
    self.commenters = ''
    self.wordchars += '%&()+-./=@[]^`{}~!'
  def dict(self):
    ret = dict()
    expect = 'name'
    curkey = None
    while True:
      token = self.get_token()
      if token == self.eof:
        if expect == self.TOKEN_KEYEND:
          token = True
          if curkey.startswith('!'):
            curkey = curkey[1:].strip()
            token = False
          ret[curkey] = token
          curkey = None
          expect = self.TOKEN_VALEND
        break
      if token == self.TOKEN_VALEND and expect == self.TOKEN_KEYEND:
        token = True
        if curkey.startswith('!'):
          curkey = curkey[1:].strip()
          token = False
        ret[curkey] = token
        curkey = None
        expect = 'key'
        continue
      if ( expect in self.TOKENS or token in self.TOKENS ) and token != expect:
        raise SyntaxError('invalid template attributes %r: expected %r, received %r'
                          % (self.src, expect, token))
      if expect == 'name':
        ret['name'] = token
        expect = self.TOKEN_VALEND
        continue
      if expect == 'key':
        curkey = token
        expect  = self.TOKEN_KEYEND
        continue
      if expect == self.TOKEN_KEYEND:
        expect = 'value'
        continue
      if expect == 'value':
        ret[curkey] = token
        curkey = None
        expect = self.TOKEN_VALEND
        continue
      if expect == self.TOKEN_VALEND:
        expect = 'key'
        continue
    if expect != self.TOKEN_VALEND:
      raise SyntaxError('invalid template attributes %r: dangling specification, expected %s'
                        % (self.src, expect))
    return ret

#------------------------------------------------------------------------------
class Compiler(object):
  '''
  A `Compiler` is the main template processor and renderer for the
  `jstc` package.
  '''

  DEFAULT_COMMENTS      = '##'
  DEFAULT_FRAGMENTS     = '##!'
  DEFAULT_HERESTRING    = '__here__'

  COLLISION_ERROR       = 'error'
  COLLISION_IGNORE      = 'ignore'
  COLLISION_OVERRIDE    = 'override'

  SPACE_PRESERVE        = 'preserve'
  SPACE_TRIM            = 'trim'
  SPACE_DEDENT          = 'dedent'
  SPACE_COLLAPSE        = 'collapse'

  ROOT_AUTO             = 'auto'

  TYPE_HTMLFRAG         = 'text/html-fragment'
  TYPE_SCRIPT           = 'text/javascript'
  TYPE_JS               = 'application/javascript'
  TYPE_JSFRAG           = 'application/javascript-fragment'

  attrprefix            = 'data-template-'
  scriptattrs           = ('name', 'locale')
  scriptfmt             = six.u('<script type="{type}" {attributes}>{script}</script>')

  default_defaults      = {
                            'space'           : SPACE_COLLAPSE,
                            'inline'          : False,
                            'precompile'      : True,
                            'collision'       : COLLISION_ERROR,
                            'pass-through'    : 'id',
                            'prefix-through'  : 'name,locale',
                          }
  default_overrides     = {}

  #----------------------------------------------------------------------------
  def __init__(self,
               attrprefix       = None,
               mimetypes        = None,
               default_type     = None,
               comments         = DEFAULT_COMMENTS,
               fragments        = DEFAULT_FRAGMENTS,
               herestring       = DEFAULT_HERESTRING,
               defaults         = None,
               overrides        = None,
               scriptattrs      = None,
               *args, **kw):
    '''
    Creates a `jstc` Compiler object.

    :Parameters:

    attrprefix : str, default: 'data-'

      The string prefixed onto attributes specified non-precompiled
      templates. This is applies to all attributes except the "type"
      attribute and any listed in the "pass-through" meta-attribute.

    mimetypes : dict, default: (loaded from plugins)

      Maps template mime-types to extensions. The keys should be
      the mime-type, and the values should be a list of extensions
      (including the "."), e.g.:

      .. code:: json

        {
          "text/x-handlebars" : [ ".handlebars", ".hbs" ],
          "text/x-mustache"   : [ ".mustache" ]
        }

    default_type : str, default: (none)

      Specifies the default mimetype for a template if it cannot
      otherwise be determined. If this is not specified, an error is
      raised if any template cannot be mapped to a mimetype.

    comments : str | null, default: '##'

      Specifies a template-agnostic line-oriented comment string.
      To disable this "global" pre-template decommenting, set this
      value to null.

    fragments : str | null, default: '##!'

      Specifies a template-agnostic token that splits a template file
      into a multi-template container. To disable this "global"
      pre-template fragmenting, set this value to null.

    herestring : str | null, default: '__here__'

      Specifies a template name component that is "self-effacing", i.e.
      a template residing in the file named::

        path/to/template/__here__.hbs

      will have the name of the template name::

        path/to/template

      To disable this self-effacing template name component feature,
      set this value to null.

    scriptattrs : list(str), default: ['name', 'locale']

      Specifies the list of template attributes, if available, that
      will be prefixed with `attrprefix` and inserted into
      non-precompiled script template nodes. Note that this can be
      overridden on a per-template basis.

    defaults : dict, default: null

      Sets the default value of template attributes, i.e. if the
      template does not specify these attributes, then it will behave
      as if they had been. See :meth:`compile` for details on known
      attributes.

    overrides : dict, default: null

      Overrides template attributes. See :meth:`compile` for details
      on known attributes.
    '''
    super(Compiler, self).__init__(*args, **kw)
    self.deftype      = default_type
    self.precompilers = dict()
    for plugin in asset.plugins('jstc.engines.plugins'):
      self.precompilers[plugin.name] = plugin.handle()
    if attrprefix is not None:
      self.attrprefix = attrprefix
    if mimetypes is not None:
      self.mimetypes = mimetypes
    if scriptattrs is not None:
      self.scriptattrs = scriptattrs
    else:
      self.mimetypes = dict()
      for name, prec in self.precompilers.items():
        for ext in prec.extensions:
          self.mimetypes[name] = prec.extensions
    self.extensions = dict()
    for mimetype, exts in self.mimetypes.items():
      for ext in exts:
        if ext in self.extensions:
          log.warning('multiple jstc pre-compilers registered for extension %r' % (ext,))
        self.extensions[ext] = mimetype
    self.commenttoken  = comments
    self.fragmenttoken = fragments
    self.heretoken     = herestring
    self.decomment_cre = None
    if self.commenttoken:
      # todo: what about lines that have data before a '##'?...  the
      #       problem is that then i really need parsing of the template
      #       to determine if the '##' is part of the template or just a
      #       comment...
      self.decomment_cre = re.compile(
        '^\\s*' + re.escape(self.commenttoken) + '.*(\r\n|\r|\n|$)', re.MULTILINE)
    self.defaults = aadict(self.default_defaults).update(defaults or {})
    self.overrides = aadict(self.default_overrides).update(overrides or {})

  #----------------------------------------------------------------------------
  def render_assets(self, assets, roots=None,
                    force_inline       = None,
                    force_precompile   = None,
                    asset_filter       = None,
                    name_transform     = None,
                    template_transform = None,
                    template_filter    = None,
                    script_wrapper     = None,
    ):
    '''
    Compiles and prepares the templates in `assets` for delivery to
    remote JavaScript template processing routines.

    :Parameters:

    assets : list

      List of asset-specifications (see `asset` python package) to
      extract templates from.

    roots : list, default: null

      Specifies the "root" portion of each corresponding asset in
      `assets`. If not specified or set to ``"auto"``, then the root
      value will be "auto-magically-guestimated"... basically, it will
      assume the root begins at the wildcard for glob-patterns, for
      example with assets in ``package:path/to/dir/**.hbs``, the root
      will be ``path/to/dir``.

      The "root" is used only by the default `asset_filter`
      implementation and performs two functions: first, it filters out
      assets that do not start with the specified root, and second,
      the root prefix is removed from the asset name to create the
      template name.

    force_inline : bool, default: null

      If specified and not null, overrides all templates' ``inline``
      attribute.

    force_precompile : bool, default: null

      If specified and not null, overrides all templates' ``precompile``
      attribute.

    asset_filter : regex | callback, default: null

      If not null, `asset_filter` is used to filter which assets are
      selected for inclusion in the rendered output. If it is a
      string, it is interpreted as a regex pattern and matched against
      the asset name. Otherwise, it is used as a callback and called
      with a single argument, the asset name, and is expected to
      return a boolean result, where truthy indicates inclusion.

      The default implementation checks that the asset name does not
      end with a tilde (``~``) and no component of the asset path
      starts with a dot (``.``).

      Note that this is NOT the same as `template_filter` since there
      may be multiple templates per asset.

    name_transform : callback, default: null

      If not null, overrides the default asset filename to template
      name evaluation with a callback that is provided two arguments,
      the asset filename and the corresponding root from `roots`, and
      is expected to return a two-element tuple of (template-name,
      mimetype). The mimetype can either be the registered mime-type
      of the template engine to use, or a file extension that can be
      used to resolve to such a mime-type. For multi-template assets,
      the template-name is used as the basename for all contained
      templates, and the mimetype is used as the default for the
      contained templates, overrides the global default, and can be
      individually overridden by each contained template. The default
      implementation uses the return value from a call to
      `os.path.splitext` and stripped of the applicable root.

    template_transform : callback, default: null

      Specifies a callback that can adjust a template's content and/or
      attributes before being filtered and processed. It is called
      with two arguments, the content and the attributes, and is
      expected to return a two-elemente tuple of (content,
      attributes).

      Note that this callback is called after `name_transform`.

    template_filter : regex | callback, default: null

      Specifies a callback that is used to filter which templates are
      selected for inclusion in the rendered output. If it is a
      string, it is interpreted as a regex pattern and matched against
      the template name. Otherwise, it is used as a callback and
      called with two arguments, the content and the attributes, and
      is expected to return a boolean result, where truthy indicates
      inclusion.

      Note that this is NOT the same as `asset_filter` since there may
      be multiple templates per asset.

      Note that this callback is called after `template_transform`.

    script_wrapper : callback, default: null

      Specifies a callback that is used to wrap a pre-compiled
      template, expressed as a JavaScript code fragment, into an HTML
      fragment. The callback is called with one argument, the raw
      JavaScript code fragment. The default simply generates a
      JavaScript "script" element, e.g.:

      .. code:: html

        <script type="text/javascript">
          {JAVASCRIPT_CODE_FRAGMENT}
        </script>
    '''
    assets = morph.tolist(assets)
    roots  = morph.tolist(roots or self.ROOT_AUTO)
    roots  = roots + ( [roots[-1]] * ( len(assets) - len(roots) ) )
    overrides = aadict(self.overrides)
    if force_inline is not None:
      overrides.inline = force_inline
    if force_precompile is not None:
      overrides.precompile = force_precompile

    # asset_filter = regex | callback | null
    if asset_filter is None:
      asset_filter = lambda name, *args, **kw: not('/.' in name or name.endswith('~'))
    elif morph.isstr(asset_filter):
      asset_filter = re.compile(asset_filter)
      asset_filter = lambda name, *args, **kw: bool(asset_filter.match(name))

    # name_transform = callback | null
    if name_transform is None:
      name_transform = self._name_transform

    # template_transform = callback | null
    if template_transform is None:
      template_transform = lambda *args, **kw: args

    # template_filter = regex | callback | null
    if template_filter is None:
      template_filter = lambda *args, **kw: True
    elif morph.isstr(template_filter):
      template_filter = re.compile(template_filter)
      template_filter = lambda text, attrs, *args, **kw: bool(template_filter.match(attrs.get('name')))

    # script_wrapper = callback | null
    if script_wrapper is None:
      script_wrapper = self.script_wrap

    hooks = aadict(
      asset_filter       = asset_filter,
      name_transform     = name_transform,
      template_transform = template_transform,
      template_filter    = template_filter,
      script_wrapper     = script_wrapper,
    )

    res = dict()
    for idx, spec in enumerate(assets):
      for text, attrs in self._asset2templates(spec, roots[idx], hooks):
        attrs = aadict(self.defaults).update(attrs).update(overrides)
        text, attrs = hooks.template_transform(text, attrs)
        if not hooks.template_filter(text, attrs):
          continue
        if not attrs.type:
          attrs.type = self.deftype
        if not attrs.type:
          raise SyntaxError(
            'could not determine template type for %r (in %r)'
            % (attrs.name, spec))
        contentType, content = self._compile(text, aadict(attrs, partial=True))
        if attrs.type not in res:
          res[attrs.type] = dict()
        if attrs.name in res[attrs.type]:
          # todo: perhaps `collision` attributes should be checked in
          #       both old and new templates... eg: imagine template A
          #       and B collide, and A specifies a `collision`
          #       attribute of "override"... i.e. A is really provided
          #       as a "default" template, then many assets can
          #       override the default without each one needing to
          #       specify the `collision` attribute.
          #       ==> also to be considered then, is should the current
          #           "collision" attribute by carried over from A to B
          #           so that if there is a "C" it will inherit that???
          col = attrs.collision
          if col == Compiler.COLLISION_ERROR:
            # todo: give more context details (eg. which asset spec, which
            #       asset item filename, etc)
            raise api.TemplateCollision(
              '%r template %r is already defined' % (attrs.type, attrs.name))
          elif col == Compiler.COLLISION_IGNORE:
            log.debug(
              '%r template %r collision: ignored' % (attrs.type, attrs.name))
            continue
          elif col != Compiler.COLLISION_OVERRIDE:
            raise TypeError(
              'unknown/unimplement collision policy %r on %r template %r'
              % (col, attrs.type, attrs.name))
          log.debug(
            '%r template %r collision: override' % (attrs.type, attrs.name))
        res[attrs.type][attrs.name] = (contentType, content, attrs)

    ret = ''
    for mimetype in sorted(res.keys()):
      templates = res[mimetype]
      types = dict()
      for name in sorted(templates.keys()):
        contentType, content, attrs = templates[name]
        if contentType not in types:
          types[contentType] = []
        types[contentType].append([contentType, content, attrs])
      for contentType in sorted(types.keys()):
        parts = types[contentType]
        if contentType == self.TYPE_HTMLFRAG:
          ret += ''.join(el[1] for el in parts)
          continue
        if contentType == self.TYPE_JS:
          content = ''.join(part[1] for part in parts)
        elif contentType == self.TYPE_JSFRAG:
          contentType, content = self._assemble(mimetype, parts)
        else:
          raise TypeError(
            'unexpected compiled template content-type %r' % (contentType,))
        ret += hooks.script_wrapper(content)
    return ret

  #----------------------------------------------------------------------------
  def _name_transform(self, name, root):
    if not name.startswith(root):
      return (None, None)
    name = name[len(root):]
    return os.path.splitext(name)

  #----------------------------------------------------------------------------
  def _asset2templates(self, spec, root, hooks):
    if root == self.ROOT_AUTO:
      # todo: this feels "primitive"... it should really extract this
      #       from globre's parse result in some way...
      if '/**' in spec:
        root = spec.split('/**', 1)[0]
        # todo: this is heavily assuming "PKG:RESOURCE" format... ugh.
        if ':' in root:
          root = root.split(':', 1)[1]
      else:
        root = ''
    if root and not root.endswith('/'):
      root += '/'
    for item in asset.load(spec):
      if not hooks.asset_filter(item.name):
        continue
      ## TODO: handle file encodings!...
      for fragment in self.fragments(item.name, root, item.read(), hooks):
        if fragment is not None:
          yield fragment

  #----------------------------------------------------------------------------
  def fragments(self, filename, root, text, hooks):
    basename, mimetype = hooks.name_transform(filename, root)
    if basename is None:
      yield None
      return
    if self.heretoken:
      basename = '/'.join(x for x in basename.split('/') if x != self.heretoken)
    attrs = aadict(name=basename, type=self.extensions.get(mimetype, mimetype))
    if not self.fragmenttoken:
      yield text, attrs
      return
    # todo: how to escape "##!" at the beginning of a line?...
    tok = self.fragmenttoken
    if not text.startswith(tok) and ( '\n' + tok ) in text:
      text = text[text.index('\n' + tok) + 1:]
    if not text.startswith(tok):
      yield text, attrs
      return
    while text:
      idx = text.index('\n')
      options = aadict(TemplateAttributeParser(text[len(tok):idx]).dict())
      if not options.name:
        raise SyntaxError('missing or invalid template name')
      options.name = os.path.join(basename, options.name)
      if 'type' in options or 'type' in attrs:
        options.type = options.get('type', attrs.type)
      text = text[idx + 1:]
      if self.heretoken:
        options.name = '/'.join(x for x in options.name.split('/') if x != self.heretoken)
      idx = text.find('\n' + tok)
      if idx < 0:
        curtext = text
        text = None
      else:
        idx += 1
        curtext = text[:idx]
        text = text[idx:]
      yield curtext, options

  #----------------------------------------------------------------------------
  def compile(self, text, attrs=None):
    '''Compiles and prepares the template `text` for delivery to remote
    JavaScript template processing routines.

    :Parameters:

    text : str

      The template content.

    attrs : dict

      Specifies the compilation control parameters. Note that these
      will be adjusted based on this compiler's `defaults` and
      `overrides` constructor parameter. Note also that some of these
      attributes are used by the `render*` methods, not here.

      name : str

        The name of this template.

      type : str

        The mime-type of the template engine intended to parse this
        template.

      collision : { 'error' | 'ignore' | 'override' }, default: 'error'

        Specifies the compiler's behavior when multiple templates have
        the same computed name. The following values are supported:

        * ``error``: throws an error and aborts compilation
        * ``ignore``: only the first template is used
        * ``override``: only the last template is used

      precompile : bool, default: true

        Sets the pre-compilation setting, i.e. whether or not a
        template should be compiled to "bytecode" (or whatever
        template-specific equivalent exists, if any) instead of being
        delivered as-is (i.e. "raw").

        NOTE: precompiled isn't *always* better and/or faster... for
        some templating languages (such as Handlebars), the
        pre-compiled template can actually be a *much* larger payload
        (in some cases 4x!) -- your mileage may vary!

      space : { 'preserve' | 'trim' | 'dedent' | 'collapse' }, default: 'collapse'

        Controls whitespace handing in template content. The following
        values are supported:

        * ``preserve``:

          Leave all whitespace exactly as-is.

        * ``trim``:

          Remove leading and trailing whitespace.

        * ``dedent``:

          "Dedent" the template (i.e. remove all whitespace that prefixes
          every line in the template) and also apply the ``trim``
          transformation.

        * ``collapse`` (the default):

          This applies the ``dedent`` transformation and then removes
          "ignorable" whitespace. Note that what is considered "ignorable"
          is dependent on the ``type``, but all assume that HTML is the
          target output. For example, for a `Handlebars`_ template, the
          following content:

          .. code::

            {{#if value}}
              <b>
                {{value}}
              </b>
            {{else}}
              <i>default</i>
            {{/if}}

          will be collapsed to:

            {{#if value}}<b>{{value}}</b>{{else}}<i>default</i>{{/if}}

      inline : bool, default: false

        Specifies whether or not this template should be rendered as
        an inlined template. Typically, inlined templates are inserted
        directly into the base HTML page, and non-inlined templates
        are loaded asynchronously.

        NOTE: this attribute is NOT used by the `jstc` package...
        this is an attribute that is intended to be used by callers to
        filter templates in `template_filter` and/or specially process
        the resulting templates.

      protected : bool, default: false

        Specifies whether or not this template should only be served
        to authenticated and authorized clients. Typically, these are
        not inlined into the base HTML page and are delivered
        asynchronously after authentication.

        NOTE: this attribute is NOT used by the `jstc` package...
        this is an attribute that is intended to be used by callers to
        filter templates in `template_filter` and/or specially process
        the resulting templates.

      partial : bool, default: false

        Specifies whether or not the return value can be in
        ``application/javascript-fragment`` format (see Returns for
        details).

      pass-through : str, default: "id"

        A comma-separated list of attributes that should be
        passed-through as-is from the `attrs` object into the
        "<script>" tag as HTML attributes. For example, if `attrs`
        contains the key/value pairs ``{'id': 'foo', 'pass-through':
        'id'}``, then the resulting "<script>" tag will be::

          <script ... id="foo">...</script>

        Note that this only applies to non-pre-compiled templates.

      prefix-through : str, default: "name,locale"

        Similar to the `pass-through` attribute, except these
        attribute names have the compiler's `attrprefix` setting
        prefixed to the HTML attribute name. For example, if `attrs`
        contains the key/value pairs ``{'media': 'print',
        'prefix-through': 'id'}`` and `attrprefix` is the default,
        then the resulting "<script>" tag will be::

          <script ... data-media="print">...</script>

        Note that this only applies to non-pre-compiled templates.

      trim : bool, default: null

        DEPRECATED; use `space`.

    :Returns:

    list

      The return value is a two-element tuple: (contentType, content).
      The `content` is the result of the compiled `text`, and contentType
      is one of the following values:

      * ``text/html-fragment``:

        For template types that do not support pre-compilation or the
        compiler failed for some other reason, the raw template is
        returned wrapped in an appropriate HTML element.

      * ``application/javascript``:

        The `content` is a pre-compiled version of `text`, suitably
        wrapped in an appropriate HTML "script" element.

      * ``application/javascript-fragment``:

        The `content` is a partially pre-compiled version of
        `text`. This means that an additional "assemble" step that
        completes the pre-compilation process is required before the
        template becomes usable. This contentType will only be
        returned if the `partial` attribute was truthy.

    '''
    attrs = aadict(self.defaults).update(attrs).update(self.overrides)
    return self._compile(text, attrs)

  #----------------------------------------------------------------------------
  def whitespace(self, text, attrs):
    xform = attrs.space
    if xform == self.SPACE_PRESERVE:
      return text
    if xform in (self.SPACE_COLLAPSE, self.SPACE_DEDENT):
      text = textwrap.dedent(text)
    if xform in (self.SPACE_COLLAPSE, self.SPACE_DEDENT, self.SPACE_TRIM):
      text = text.strip()
    if xform in (self.SPACE_COLLAPSE,):
      precomp = self.precompilers.get(attrs.type)
      if not precomp:
        # todo: is this the right way to default the whitespace
        #       collapsing engine?
        from .engines.base import Engine
        precomp = Engine()
      text = precomp.whitespace(text, attrs)
    return text


  #----------------------------------------------------------------------------
  def _compile(self, text, attrs):
    # todo: should `attrs` be able to specify an alternate commenting scheme?...
    text = self.decomment(text)
    xform = attrs.space
    if attrs.trim is False:
      xform = self.SPACE_PRESERVE
    text = self.whitespace(text, aadict(attrs, space=xform))
    if not morph.tobool(attrs.precompile):
      # return (text, False, attrs)
      return self._htmlwrap(text, attrs)
    precomp = self.precompilers.get(attrs.type)
    if not precomp or precomp.precompile is api.PrecompilerUnavailable:
      return self._htmlwrap(text, attrs)
    try:
      return precomp.precompile(text, attrs=attrs)
    except api.PrecompilerUnavailable:
      return self._htmlwrap(text, attrs)

  #----------------------------------------------------------------------------
  def _assemble(self, mimetype, parts):
    precomp = self.precompilers.get(mimetype)
    if not precomp or precomp.assemble is api.PrecompilerUnavailable:
      raise ValueError(
        '%r template engine not registered or does not support `assemble`',
        mimetype)
    return precomp.assemble(parts)

  #----------------------------------------------------------------------------
  def _htmlwrap(self, text, attrs):
    pfx = self.attrprefix or ''
    curattrs = dict()
    for key in [k.strip() for k in attrs.get('prefix-through', '').split(',')]:
      if key and key in attrs:
        curattrs[pfx + key] = attrs.get(key)
    for key in [k.strip() for k in attrs.get('pass-through', '').split(',')]:
      if key and key in attrs:
        curattrs[key] = attrs.get(key)
    return (
      self.TYPE_HTMLFRAG,
      self.script_wrap(text, type=attrs.type, attrs=curattrs),
    )

  #----------------------------------------------------------------------------
  def script_wrap(self, script, type=None, attrs=None):
    attrtext = ' '.join(
      '{key}="{value}"'.format(
        key=key, value=cgi.escape(attrs[key], quote=True))
      for key in sorted((attrs or {}).keys())
    )
    return self.scriptfmt.format(
      type       = type or self.TYPE_SCRIPT,
      attributes = attrtext,
      script     = self.script_escape(script),
    )

  #----------------------------------------------------------------------------
  def script_escape(self, text):
    # todo: do i also need to escape '<', '>', and '&'?...
    text = text.replace('<![CDATA', '<![CDATA[<]]>![CDATA')
    return text.replace('</script', '<![CDATA[<]]>/script')

  #----------------------------------------------------------------------------
  def decomment(self, text):
    '''
    Removes template-agnostic global comment lines from `text`.
    '''
    if self.decomment_cre is None:
      return text
    return self.decomment_cre.sub('', text)


#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2016/09/15
# copy: (C) Copyright 2016-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import unittest
import os
import textwrap

from aadict import aadict
import fso

#------------------------------------------------------------------------------
class TestCompiler(unittest.TestCase):

  maxDiff = None

  #----------------------------------------------------------------------------
  def test_fragments(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler()
    hooks = aadict(name_transform=compiler._name_transform)
    self.assertEqual(
      list(compiler.fragments('foo/bar.jst', '', 'i am a template.', hooks)),
      [('i am a template.', aadict(name='foo/bar', type='.jst'))])
    self.assertEqual(
      list(compiler.fragments('foo/bar.jst', '', '''\
##! zig
  i am the zig template.
##! __here__
  i am the root template.
''', hooks)),
      [
        ('  i am the zig template.\n',  aadict(name='foo/bar/zig', type='.jst')),
        ('  i am the root template.\n', aadict(name='foo/bar', type='.jst')),
      ])

  #----------------------------------------------------------------------------
  def test_attributes(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler()
    hooks = aadict(name_transform=compiler._name_transform)
    self.assertEqual(
      list(compiler.fragments('foo/bar.jst', '', '''\
##! zig; channels: "public,protected"
  i am the zig template.
##! __here__; public; protected
  i am the root template.
##! zag; type: text/jst; !public; !protected
  i am the zag template.
''', hooks)),
      [
        ('  i am the zig template.\n',  aadict(name='foo/bar/zig', type='.jst', channels='public,protected')),
        ('  i am the root template.\n', aadict(name='foo/bar', type='.jst', public=True, protected=True)),
        ('  i am the zag template.\n',  aadict(name='foo/bar/zag', type='text/jst', public=False, protected=False)),
      ])

  #----------------------------------------------------------------------------
  def writecontent(self, files, dedent=True):
    for name, content in files.items():
      path = os.path.join(os.path.dirname(__file__), name)
      pdir = os.path.dirname(path)
      if not os.path.isdir(pdir):
        os.makedirs(pdir)
      with open(path, 'wb') as fp:
        fp.write(textwrap.dedent(content))

  #----------------------------------------------------------------------------
  def test_render_simple(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(
      overrides=dict(inline=True, precompile=False))
    with fso.push() as overlay:
      self.writecontent({
        'test/common/hello.hbs':
          '''\
            ##! __here__
              Hello, world!
            ##! name
              Hello, {{name}}!
          '''
      })
      self.assertEqual(
        compiler.render_assets('jstc:test/common/hello.hbs', 'test'),
        '''\
<script type="text/x-handlebars" data-template-name="common/hello">Hello, world!</script>\
<script type="text/x-handlebars" data-template-name="common/hello/name">Hello, {{name}}!</script>\
''')

  #----------------------------------------------------------------------------
  def test_render_trim_deprecated(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(
      overrides=dict(inline=True, precompile=False))
    with fso.push() as overlay:
      self.writecontent({
        'test.hbs':
          '''\
            ##! 0-default
              <span>
                text
              </span>
            ##! 1-trim; trim
              <span>
                text
              </span>
            ##! 2-notrim; !trim
              <span>
                text
              </span>
          '''
      })
      self.assertEqual(
        compiler.render_assets('jstc:test.hbs'),
        '''\
<script type="text/x-handlebars" data-template-name="test/0-default"><span>
  text
</span></script>\
<script type="text/x-handlebars" data-template-name="test/1-trim"><span>
  text
</span></script>\
<script type="text/x-handlebars" data-template-name="test/2-notrim">  <span>
    text
  </span>
</script>\
''')

  #----------------------------------------------------------------------------
  def test_render_space(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(
      overrides=dict(inline=True, precompile=False))
    with fso.push() as overlay:
      self.writecontent({
        'test.hbs':
          '''\
            ##! 0-default
              {{#if value}}
                <span>
                  {{value}}
                </span>
              {{else}}
                <span>default</span>
              {{/if}}
            ##! 1-preserve; space: preserve
              {{#if value}}
                <span>
                  {{value}}
                </span>
              {{else}}
                <span>default</span>
              {{/if}}
            ##! 2-trim; space: trim
              {{#if value}}
                <span>
                  {{value}}
                </span>
              {{else}}
                <span>default</span>
              {{/if}}
            ##! 3-dedent; space: dedent
              {{#if value}}
                <span>
                  {{value}}
                </span>
              {{else}}
                <span>default</span>
              {{/if}}
            ##! 4-collapse; space: collapse
              {{#if value}}
                <span>
                  {{value}}
                </span>
              {{else}}
                <span>default</span>
              {{/if}}
          '''
      })
      self.assertEqual(
        compiler.render_assets('jstc:test.hbs'),
        '''\
<script type="text/x-handlebars" data-template-name="test/0-default">{{#if value}}<span>{{value}}</span>{{else}}<span>default</span>{{/if}}</script>\
<script type="text/x-handlebars" data-template-name="test/1-preserve">  {{#if value}}
    <span>
      {{value}}
    </span>
  {{else}}
    <span>default</span>
  {{/if}}
</script>\
<script type="text/x-handlebars" data-template-name="test/2-trim">{{#if value}}
    <span>
      {{value}}
    </span>
  {{else}}
    <span>default</span>
  {{/if}}</script>\
<script type="text/x-handlebars" data-template-name="test/3-dedent">{{#if value}}
  <span>
    {{value}}
  </span>
{{else}}
  <span>default</span>
{{/if}}</script>\
<script type="text/x-handlebars" data-template-name="test/4-collapse">{{#if value}}<span>{{value}}</span>{{else}}<span>default</span>{{/if}}</script>\
''')


#   #----------------------------------------------------------------------------
#   def test_render_unicode(self):
#     import jstc.compiler
#     compiler = jstc.compiler.Compiler(
#       overrides=dict(inline=True, precompile=False))
#     with fso.push() as overlay:
#       self.writecontent({
#         'test/temp.hbs':
#           '''\
#             ##! __here__
#               It's {{ value }}°C, et voilà, 吹!
#           '''
#       })
#       self.assertEqual(
#         compiler.render_assets('jstc:test/temp.hbs', 'test'),
#         '''\
# <script type="text/x-handlebars" data-template-name="temp">It's {{ value }}°C, et voilà, 吹!</script>\
# ''')

  #----------------------------------------------------------------------------
  def test_comments(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(
      overrides=dict(inline=True, precompile=False))
    with fso.push() as overlay:
      self.writecontent({
        'test/application.hbs':
          '''\
            <div>
              ## TODO: super-secret comment!
              Nothing to see here.
            </div>
          '''
      })
      self.assertEqual(
        compiler.render_assets('jstc:test/application.hbs', 'test'),
        '''\
<script type="text/x-handlebars" data-template-name="application"><div>
  Nothing to see here.
</div>\
</script>\
''')

  #----------------------------------------------------------------------------
  def test_root(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(
      overrides=dict(inline=True, precompile=False))
    with fso.push() as overlay:
      self.writecontent({
        'test/one/template.hbs': 'template "one".',
        'test/two/template.hbs': 'template "two".',
      })
      self.assertEqual(
        compiler.render_assets('jstc:test/one/template.hbs', 'test/one'),
        '''\
<script type="text/x-handlebars" data-template-name="template">template "one".</script>\
''')
      self.assertEqual(
        compiler.render_assets('jstc:test/two/template.hbs', 'test/two'),
        '''\
<script type="text/x-handlebars" data-template-name="template">template "two".</script>\
''')
      self.assertEqual(
        compiler.render_assets(
          ['jstc:test/one/template.hbs', 'jstc:test/two/template.hbs'], 'test'),
        '''\
<script type="text/x-handlebars" data-template-name="one/template">template "one".</script>\
<script type="text/x-handlebars" data-template-name="two/template">template "two".</script>\
''')

  #----------------------------------------------------------------------------
  def test_collision_error(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(
      overrides=dict(inline=True, precompile=False))
    with fso.push() as overlay:
      self.writecontent({
        'test/one/template.hbs': 'template "one".',
        'test/two/template.hbs': 'template "two".',
      })
      with self.assertRaises(jstc.TemplateCollision) as cm:
        compiler.render_assets(
          ['jstc:test/one/template.hbs', 'jstc:test/two/template.hbs'],
          ['test/one', 'test/two'])
      self.assertEqual(
        str(cm.exception),
        ''''text/x-handlebars' template 'template' is already defined''')

  #----------------------------------------------------------------------------
  def test_collision_ignore(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(
      defaults=dict(collision='ignore'),
      overrides=dict(inline=True, precompile=False),
    )
    with fso.push() as overlay:
      self.writecontent({
        'test/one/template.hbs': 'template "one".',
        'test/two/template.hbs': 'template "two".',
      })
      self.assertEqual(
        compiler.render_assets(
          ['jstc:test/one/template.hbs', 'jstc:test/two/template.hbs'],
          ['test/one', 'test/two']),
        '''\
<script type="text/x-handlebars" data-template-name="template">template "one".</script>\
''')

  #----------------------------------------------------------------------------
  def test_collision_override(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(
      defaults=dict(collision='override'),
      overrides=dict(inline=True, precompile=False),
    )
    with fso.push() as overlay:
      self.writecontent({
        'test/one/template.hbs': 'template "one".',
        'test/two/template.hbs': 'template "two".',
      })
      self.assertEqual(
        compiler.render_assets(
          ['jstc:test/one/template.hbs', 'jstc:test/two/template.hbs'],
          ['test/one', 'test/two']),
        '''\
<script type="text/x-handlebars" data-template-name="template">template "two".</script>\
''')

  #----------------------------------------------------------------------------
  def test_collision_pertemplate(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(
      defaults=dict(collision='ignore'),
      overrides=dict(inline=True, precompile=False),
    )
    with fso.push() as overlay:
      self.writecontent({
        'test/one/template.hbs':
          '''\
            ##! a
              template "one/a".
            ##! b
              template "one/b".
          ''',
        'test/two/template.hbs':
          '''\
            ##! a; collision: ignore
              template "two/a".
            ##! b; collision: override
              template "two/b".
          ''',
      })
      self.assertEqual(
        compiler.render_assets(
          ['jstc:test/one/template.hbs', 'jstc:test/two/template.hbs'],
          ['test/one', 'test/two']),
        '''\
<script type="text/x-handlebars" data-template-name="template/a">template "one/a".</script>\
<script type="text/x-handlebars" data-template-name="template/b">template "two/b".</script>\
''')

  #----------------------------------------------------------------------------
  def test_precompile(self):
    import jstc
    with fso.push() as overlay:
      self.writecontent({
        'test/hello.hbs': 'hello, world!',
        'test/hello/name.hbs': 'hello, {{name}}!',
      })
      compiled = jstc.render_assets('jstc:test/**.hbs', force_inline=True, force_precompile=True)
      if 'text/x-handlebars' in compiled:
        raise unittest.SkipTest(
          'handlebars executable not available (use "npm install handlebars")')
      self.assertMultiLineEqual(
        compiled,
        '''\
<script type="text/javascript" >(function(){var t=Handlebars.template,ts=Handlebars.templates=Handlebars.templates||{};ts["hello"]=t({"compiler":[7,">= 4.0.0"],"main":function(container,depth0,helpers,partials,data) {
    return "hello, world!";
},"useData":true});ts["hello/name"]=t({"compiler":[7,">= 4.0.0"],"main":function(container,depth0,helpers,partials,data) {
    var helper;

  return "hello, "
    + container.escapeExpression(((helper = (helper = helpers.name || (depth0 != null ? depth0.name : depth0)) != null ? helper : helpers.helperMissing),(typeof helper === "function" ? helper.call(depth0 != null ? depth0 : {},{"name":"name","hash":{},"data":data}) : helper)))
    + "!";
},"useData":true});})();</script>''')

  #----------------------------------------------------------------------------
  def test_asset_filter(self):
    import jstc
    with fso.push() as overlay:
      self.writecontent({
        'test/hello.hbs': 'hello!',
        'test/goodbye.hbs': 'so long!',
      })
      self.assertEqual(
        jstc.render_assets('jstc:test/**.hbs', force_inline=True, force_precompile=False),
        '''\
<script type="text/x-handlebars" data-template-name="goodbye">so long!</script>\
<script type="text/x-handlebars" data-template-name="hello">hello!</script>\
''')
      self.assertEqual(
        jstc.render_assets(
          'jstc:test/**.hbs', force_inline=True, force_precompile=False,
          asset_filter=lambda name: name == 'test/hello.hbs'),
        '''\
<script type="text/x-handlebars" data-template-name="hello">hello!</script>\
''')
      self.assertEqual(
        jstc.render_assets('jstc:test/**.hbs', force_inline=True, force_precompile=False,
          asset_filter=lambda name: name != 'test/hello.hbs'),
        '''\
<script type="text/x-handlebars" data-template-name="goodbye">so long!</script>\
''')

  #----------------------------------------------------------------------------
  def test_name_transform(self):
    import jstc
    with fso.push() as overlay:
      self.writecontent({
        'test/hello.hbs': 'hello!',
        'test/goodbye.hbs': 'so long!',
      })
      def mynt(name, root):
        return (name[2:].replace('d', 'd-').split('.')[0], 'text/x-mustache')
      self.assertEqual(
        jstc.render_assets('jstc:test/**.hbs', force_inline=True, force_precompile=False,
          name_transform=mynt),
        '''\
<script type="text/x-mustache" data-template-name="st/good-bye">so long!</script>\
<script type="text/x-mustache" data-template-name="st/hello">hello!</script>\
''')

  #----------------------------------------------------------------------------
  def test_template_transform(self):
    import jstc
    with fso.push() as overlay:
      self.writecontent({
        'test/hello.hbs': 'hello!',
        'test/goodbye.hbs': 'so long!',
      })
      def mytt(text, attrs):
        if attrs.name == 'hello':
          text = 'hello, world!'
          attrs.id = 'HW'
        else:
          attrs.type = 'template/jst'
        return (text, attrs)
      self.assertEqual(
        jstc.render_assets('jstc:test/**.hbs', force_inline=True, force_precompile=False,
          template_transform=mytt),
        '''\
<script type="template/jst" data-template-name="goodbye">so long!</script>\
<script type="text/x-handlebars" data-template-name="hello" id="HW">hello, world!</script>\
''')

  #----------------------------------------------------------------------------
  def test_template_filter(self):
    import jstc
    with fso.push() as overlay:
      self.writecontent({
        'test/hello.hbs': 'hello!',
        'test/goodbye.hbs': '''\
##! __here__
  so long!
##! friend
  ciao!
'''
      })
      self.assertEqual(
        jstc.render_assets('jstc:test/**.hbs', force_inline=True, force_precompile=False),
        '''\
<script type="text/x-handlebars" data-template-name="goodbye">so long!</script>\
<script type="text/x-handlebars" data-template-name="goodbye/friend">ciao!</script>\
<script type="text/x-handlebars" data-template-name="hello">hello!</script>\
''')
      self.assertEqual(
        jstc.render_assets('jstc:test/**.hbs', force_inline=True, force_precompile=False,
          template_filter=lambda text, attrs: 'ciao' not in text),
        '''\
<script type="text/x-handlebars" data-template-name="goodbye">so long!</script>\
<script type="text/x-handlebars" data-template-name="hello">hello!</script>\
''')


  #----------------------------------------------------------------------------
  def test_script_wrapper(self):
    import jstc
    with fso.push() as overlay:
      self.writecontent({
        'test/hello.hbs': 'hello, world!',
        'test/hello/name.hbs': 'hello, {{name}}!',
      })
      compiled = jstc.render_assets(
        'jstc:test/**.hbs', force_inline=True, force_precompile=True,
        script_wrapper = lambda script, *args, **kw: '<SCRIPT>' + script + '</SCRIPT>')
      if 'text/x-handlebars' in compiled:
        raise unittest.SkipTest(
          'handlebars executable not available (use "npm install handlebars")')
      self.assertMultiLineEqual(
        compiled,
        '''\
<SCRIPT>(function(){var t=Handlebars.template,ts=Handlebars.templates=Handlebars.templates||{};ts["hello"]=t({"compiler":[7,">= 4.0.0"],"main":function(container,depth0,helpers,partials,data) {
    return "hello, world!";
},"useData":true});ts["hello/name"]=t({"compiler":[7,">= 4.0.0"],"main":function(container,depth0,helpers,partials,data) {
    var helper;

  return "hello, "
    + container.escapeExpression(((helper = (helper = helpers.name || (depth0 != null ? depth0.name : depth0)) != null ? helper : helpers.helperMissing),(typeof helper === "function" ? helper.call(depth0 != null ? depth0 : {},{"name":"name","hash":{},"data":data}) : helper)))
    + "!";
},"useData":true});})();</SCRIPT>''')

#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2016/09/15
# copy: (C) Copyright 2016-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import unittest

from .. import api

#------------------------------------------------------------------------------
class TestHandlebarsEngine(unittest.TestCase):

  #----------------------------------------------------------------------------
  def test_partial(self):
    try:
      from .handlebars import HandlebarsEngine
      # todo: make the compiler version details regex'ed...
      self.assertEqual(
        HandlebarsEngine().precompile('hello, world!', {'name': 'foo', 'partial': True}),
        ('application/javascript-fragment',
         '''\
{"compiler":[7,">= 4.0.0"],"main":function(container,depth0,helpers,partials,data) {
    return "hello, world!";
},"useData":true}'''))
    except api.PrecompilerUnavailable:
      raise unittest.SkipTest(
        'handlebars executable not available (use "npm install handlebars")')

  #----------------------------------------------------------------------------
  def test_assemble(self):
    try:
      from .handlebars import HandlebarsEngine
      comp = HandlebarsEngine()
      parts = []
      for name, text in [
          ('hello',      'hello, world!'),
          ('hello/name', 'hello, {{name}}!'),
        ]:
        attrs = dict(name=name, partial=True)
        parts.append(comp.precompile(text, attrs) + (attrs,))
      # todo: make the compiler version details regex'ed...
      self.assertEqual(
        comp.assemble(parts),
        ('application/javascript',
         '''\
(function(){var t=Handlebars.template,ts=Handlebars.templates=Handlebars.templates||{};ts["hello"]=t({"compiler":[7,">= 4.0.0"],"main":function(container,depth0,helpers,partials,data) {
    return "hello, world!";
},"useData":true});ts["hello/name"]=t({"compiler":[7,">= 4.0.0"],"main":function(container,depth0,helpers,partials,data) {
    var helper;

  return "hello, "
    + container.escapeExpression(((helper = (helper = helpers.name || (depth0 != null ? depth0.name : depth0)) != null ? helper : helpers.helperMissing),(typeof helper === "function" ? helper.call(depth0 != null ? depth0 : {},{"name":"name","hash":{},"data":data}) : helper)))
    + "!";
},"useData":true});})();'''))
    except api.PrecompilerUnavailable:
      raise unittest.SkipTest(
        'handlebars executable not available (use "npm install handlebars")')

  #----------------------------------------------------------------------------
  def test_standalone(self):
    try:
      from .handlebars import HandlebarsEngine
      # todo: make the compiler version details regex'ed...
      self.assertEqual(
        HandlebarsEngine().precompile('this is {{name}}', {'name': 'foo'}),
        ('application/javascript',
         '''\
(function(){var t=Handlebars.template,ts=Handlebars.templates=Handlebars.templates||{};ts["foo"]=t({"compiler":[7,">= 4.0.0"],"main":function(container,depth0,helpers,partials,data) {
    var helper;

  return "this is "
    + container.escapeExpression(((helper = (helper = helpers.name || (depth0 != null ? depth0.name : depth0)) != null ? helper : helpers.helperMissing),(typeof helper === "function" ? helper.call(depth0 != null ? depth0 : {},{"name":"name","hash":{},"data":data}) : helper)));
},"useData":true});})();'''))
    except api.PrecompilerUnavailable:
      raise unittest.SkipTest(
        'handlebars executable not available (use "npm install handlebars")')


#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2016/09/12
# copy: (C) Copyright 2016-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import os, sys, setuptools
from setuptools import setup, find_packages

# require python 2.7+
if sys.hexversion < 0x02070000:
  raise RuntimeError('This package requires python 2.7 or better')

heredir = os.path.abspath(os.path.dirname(__file__))
def read(*parts, **kw):
  try:    return open(os.path.join(heredir, *parts)).read()
  except: return kw.get('default', '')

test_dependencies = [
  'nose                 >= 1.3.0',
  'coverage             >= 3.5.3',
  'fso                  >= 0.3.1',
]

dependencies = [
  'asset                >= 0.6.10',
  'six                  >= 1.6.0',
  'morph                >= 0.1.2',
]

entrypoints = {
  ## TODO: use `asset.find_plugins` when it exists...
  'jstc.engines.plugins' : [
    'text/x-handlebars  = jstc.engines.handlebars:HandlebarsEngine',
    'text/x-mustache    = jstc.engines.mustache:MustacheEngine',
  ],
}

classifiers = [
  'Intended Audience :: Developers',
  'Programming Language :: Python',
  'Operating System :: OS Independent',
  'Topic :: Internet',
  'Topic :: Software Development',
  'Topic :: Internet :: WWW/HTTP',
  'Natural Language :: English',
  'License :: OSI Approved :: MIT License',
  'License :: Public Domain',
]

setup(
  name                  = 'jstc',
  version               = read('VERSION.txt', default='0.0.1').strip(),
  description           = 'Server-side JavaScript template compilation and packaging',
  long_description      = read('README.rst'),
  classifiers           = classifiers,
  author                = 'Philip J Grabner, Canary Health Inc',
  author_email          = 'oss@canary.md',
  url                   = 'http://github.com/canaryhealth/jstc',
  keywords              = 'javascript server-side template compiler packager',
  packages              = find_packages(),
  include_package_data  = True,
  zip_safe              = True,
  install_requires      = dependencies,
  tests_require         = test_dependencies,
  test_suite            = 'jstc',
  entry_points          = entrypoints,
  license               = 'MIT (http://opensource.org/licenses/MIT)',
)

#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2016/09/15
# copy: (C) Copyright 2016-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

'''
The `jstc` Python package compiles and packages JavaScript templates
for delivery to browsers for client-side evaluation.

See `LICENSE.txt` for licensing details.
'''

from .compiler import Compiler
from .api import *

#------------------------------------------------------------------------------
_default_compiler = None

#------------------------------------------------------------------------------
def get_default_compiler():
  '''
  Returns the global default `jstc.compiler.Compiler` singleton used
  by the `jstc.compile` and `jstc.render*` functions.
  '''
  global _default_compiler
  if _default_compiler is None:
    _default_compiler = Compiler()
  return _default_compiler

#------------------------------------------------------------------------------
def set_default_compiler(compiler):
  '''
  Sets the global default `jstc.compiler.Compiler` singleton used by
  the `jstc.compile` and `jstc.render*` functions.
  '''
  global _default_compiler
  _default_compiler = compiler
  return _default_compiler

#------------------------------------------------------------------------------
for method in ('compile', 'render_assets'):
  # todo: copy over method args/params signature so that ``help(jstc)``
  #       reports that correctly...
  meth = getattr(Compiler, method)
  def _makefunc(method):
    def _func(*args, **kw):
      return getattr(get_default_compiler(), method)(*args, **kw)
    return _func
  _func = _makefunc(method)
  for attr in ('__doc__', '__name__'):
    setattr(_func, attr, getattr(meth, attr))
  locals()[method] = _func
del method, attr

#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2016/09/16
# copy: (C) Copyright 2016-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import unittest
import fso
import os
import textwrap

from .. import api

#------------------------------------------------------------------------------
class TestMustacheEngine(unittest.TestCase):

  #----------------------------------------------------------------------------
  def writecontent(self, files, dedent=True):
    for name, content in files.items():
      path = os.path.join(os.path.dirname(__file__), name)
      pdir = os.path.dirname(path)
      if not os.path.isdir(pdir):
        os.makedirs(pdir)
      with open(path, 'wb') as fp:
        fp.write(textwrap.dedent(content))

  #----------------------------------------------------------------------------
  def test_direct(self):
    from .mustache import MustacheEngine
    engine = MustacheEngine()
    self.assertIs(engine.precompile, api.PrecompilerUnavailable)

  #----------------------------------------------------------------------------
  def test_compiler(self):
    import jstc.compiler
    compiler = jstc.compiler.Compiler(defaults=dict(inline=True, precompile=True))
    with fso.push() as overlay:
      self.writecontent({
        'test/template.mustache': 'A mustache template.',
      })
      self.assertEqual(
        compiler.render_assets(
          'jstc:engines/test/template.mustache', 'engines/test'),
        '<script type="text/x-mustache" data-template-name="template">A mustache template.</script>'
      )


#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------
